         .586p
DGROUP group _DATA,_BSS
_TEXT  segment para public use32 'CODE' 
       assume cs:_TEXT, ds:DGROUP, ss:DGROUP 

public soundstart_
soundstart_:                            ; Used to check code alignment
startlbl:

;---------------------------
; Petter Schau 1996,97
; Sound emulation for Fellow 
; Soundcard independent part
;---------------------------


include defs.inc
include func\memory.inc
include func\inout.inc



#ifndef SOUND_C
;-------------------
; Audio IO Registers
;-------------------
; Shared stubs for all 4 channels, the channel number is extracted
; from ecx (destination address)

GET_CHANNEL_NUMBER macro
                and     ecx,070h
                shr     ecx,4
                sub     ecx,2
endm


;=======
; AUDXPT
;=======
; $dff0a0,b0,c0,d0


                ALIGN32

public waudXpth_
waudXpth_:      GET_CHANNEL_NUMBER
                and     edx,01fh                        ; Store 21 bit (2 MB)
                mov     word ptr [_audpt + 2 + 4*ecx],dx
                ret

; $dff0a2,b2,c2,d2

                ALIGN32

public waudXptl_
waudXptl_:      GET_CHANNEL_NUMBER
                and     edx,0fffffffeh                  ; No odd adresses
                mov     word ptr [_audpt + 4*ecx],dx
                ret

;========
; AUDXLEN
;========
; $dff0a4,b4,c4,d4

                ALIGN32

public waudXlen_
waudXlen_:      GET_CHANNEL_NUMBER
                and     edx,0ffffh
                mov     dword ptr [_audlen + 4*ecx],edx
                ret

;========
; AUDXPER
;========
; $dff0a6,b6,c6,d6

                ALIGN32

public waudXper_
waudXper_:      GET_CHANNEL_NUMBER
                and     edx,0ffffh
                mov     edx,dword ptr [_periodtable + 4*edx]
                mov     dword ptr [_audper + 4*ecx],edx
                ret

;========
; AUDXVOL
;========
; $dff0a8,b8,c8,d8

; Replay routines access volume as a byte register sometimes...

                ALIGN32

public waudXvol_
waudXvol_:      GET_CHANNEL_NUMBER
                test    dl,dl           ; Kludge, if lo is zero and hi data
                jnz     nokludge        ; switch, because byte access likely
                test    dh,dh
                jz      nokludge
                mov     dl,dh
nokludge:       and     edx,0ffh
                test    edx,64
                jz      audXvolok
                mov     edx,63
audXvolok:      and     edx,3fh
                shl     edx,1
                mov     dword ptr [_audvol + 4*ecx],edx
                ret

;===========================
; Audio state machine
; IN:  edx - channel*4
; OUT: ecx and edx - garbage
;===========================


                        ALIGN32

public audiostate0_
audiostate0_:
                        ;-------------------
                        ; Statechange 0 to 1
                        ;-------------------

                        mov     ecx,[_audlen+edx]
                        mov     [_audlenw+edx],ecx
                        mov     ecx,[_audpt+edx]
                        mov     [_audptw+edx],ecx
                        mov     dword ptr [_audpercounter+edx],0
                        mov     dword ptr [_audstate+edx],offset audiostate1_
                        ret


                        ALIGN32

;---------
; State 1
;---------


public audiostate1_
audiostate1_:
                        ;-------------------
                        ; Statechange 1 to 5
                        ;-------------------

as15:                   cmp     dword ptr [_audlenw+edx],1 
                        je      as15a
                        dec     dword ptr [_audlenw+edx]
as15a:                  mov     dword ptr [_audstate+edx],offset audiostate5_
                        mov     edx,[_audioirqmask+edx]
                        mov     ecx,0138h                          ; 09c * 2
                        or      dh,080h
                        jmp     dword ptr [_iowrite+ecx]

;---------
; State 2
;---------

                        ALIGN32

public audiostate2_
audiostate2_:           cmp     dword ptr [_audpercounter+edx],010000h
                        jb      as22

                        ;-------------------
                        ; Statechange 2 to 3
                        ;-------------------

                        sub     dword ptr [_audpercounter+edx],010000h
                        mov     ecx,[_audper+edx]
                        add     [_audpercounter+edx],ecx
                        mov     ecx,[_audvol+edx]
                        mov     [_audvolw+edx],ecx
                        mov     dword ptr [_audstate+edx],offset audiostate3_
                        xor     ecx,ecx                        
                        mov     cl,byte ptr [_auddat+edx]
                        shl     ecx,7
                        add     ecx,[_audvolw + edx]
                        mov     cx,[_volumes + ecx]
                        mov     word ptr [_auddatw + edx],cx
                        ret

                        ;-------------------
                        ; Statechange 2 to 2
                        ;-------------------

as22:                   mov     ecx,[_audper+edx]
                        add     [_audpercounter+edx],ecx
                        ret

;---------
; State 3
;---------

                        ALIGN32

public audiostate3_
audiostate3_:           cmp     dword ptr [_audpercounter+edx],010000h
                        jb      as33

                        ;-------------------
                        ; Statechange 3 to 2
                        ;-------------------

                        sub     dword ptr [_audpercounter+edx],010000h
                        mov     ecx,[_audper[edx]
                        add     [_audpercounter+edx],ecx
                        mov     ecx,[_audvol+edx]
                        mov     dword ptr [_audvolw+edx],ecx
                        mov     dword ptr [_audstate+edx],offset audiostate2_
                        xor     ecx,ecx                        
                        mov     cl,byte ptr [_auddat + 1 + edx]
                        shl     ecx,7
                        add     ecx,[_audvolw+edx]
                        mov     cx,[_volumes+ecx]
                        mov     word ptr [_auddatw + edx],cx
                        mov     ecx,[_audptw+edx]
                        add     dword ptr [_audptw+edx],2
                        and     dword ptr [_audptw+edx],01fffffh
                        mov     cx,[_cmem+ecx]
                        mov     word ptr [_auddat + edx],cx
                        cmp     dword ptr [_audlenw+edx],1
                        je      as32a
                        dec     dword ptr [_audlenw+edx]
                        ret

as32a:                  mov     ecx,[_audlen+edx]
                        mov     [_audlenw+edx],ecx
                        mov     ecx,[_audpt+edx]
                        mov     [_audptw+edx],ecx
                        mov     edx,[_audioirqmask+edx]
                        mov     ecx,0138h                               ; 9c*2
                        or      dh,080h
                        jmp     dword ptr [_iowrite+ecx]

                        ;-------------------
                        ; Statechange 3 to 3
                        ;-------------------

as33:                   mov     ecx,[_audper+edx]
                        add     [_audpercounter+edx],ecx
                        ret

                        ALIGN32

public audiostate4_
public audiostate6_
audiostate6_:
audiostate4_:           ret



;---------
; State 5
;---------

                        ;-------------------
                        ; Statechange 5 to 2
                        ;-------------------

                        ALIGN32

public audiostate5_
audiostate5_:           mov     ecx,[_audvol+edx]
                        mov     [_audvolw+edx],ecx
                        mov     dword ptr [_audpercounter+edx],0
                        mov     ecx,[_audptw+edx]
                        add     dword ptr [_audptw+edx],2
                        and     dword ptr [_audptw+edx],01ffffeh
                        mov     cx,[_cmem+ecx]
                        mov     word ptr [_auddat + edx],cx
                        mov     dword ptr [_audstate+edx],offset audiostate2_
                        cmp     dword ptr [_audlenw+edx],1
                        je      as52a
                        dec     dword ptr [_audlenw+edx]
                        ret
as52a:                  mov     ecx,[_audlen+edx]
                        mov     [_audlenw+edx],ecx
                        mov     ecx,[_audpt+edx]
                        mov     [_audptw+edx],ecx
                        mov     edx,[_audioirqmask+edx]
                        mov     ecx,0138h                                ;9c*2
                        or      dh,080h
                        jmp     dword ptr [_iowrite+ecx]




;------------------------------------------------
; This is called by DMACON when DMA is turned off
; In: ecx - channel number
;------------------------------------------------

                        ALIGN32

public audio_kill_channel_
audio_kill_channel_:    mov     dword ptr [_auddatw + 4*ecx],0
                        mov     dword ptr [_audstate + 4*ecx],offset audiostate0_
                        ret




;-----------------------------------------------------------------
; Frequency handlers:
;
; These are called by the emulation from the end_of_line function.
; It will generate samples in two arrays called _left and _right
; These are always 16-bit samples and is the mixed output of
; the two left channels and the two right channels.
; The number of samples they produce is dependent on the output
; frequency selected for the PC soundcard.
;
; When these finish, they jump to the function pointed to by
; _audio_putinbuffer, with the number of new samples in ebx
;-----------------------------------------------------------------

; 
; 44100 hz   (Actually 45454 sample rate on SB)


                ALIGN32

public audio_frequency_handler_44100_:
audio_frequency_handler_44100_:
                xor     ebx,ebx
                mov     eax,[_audiocounter]
size441:        add     eax,90020
                inc     ebx
                cmp     eax,040000h
                jbe     size441
                sub     eax,040000h
                mov     [_audiocounter],eax

                ; Here ebx is number of samples to add

                xor     eax,eax
                mov     dword ptr [_left],eax
                mov     dword ptr [_left+2],eax
                mov     dword ptr [_right],eax
                mov     dword ptr [_right+2],eax

                test    byte ptr [_dmacon],1      ; Channel on?
                jz      channel0441off            ; No, don't make sound 
chan0441l:      xor     edx,edx
                call    dword ptr [_audstate+edx]
                mov     ecx,dword ptr [_auddatw]
                mov     word ptr [_left+2*eax],cx
                inc     eax
                cmp     eax,ebx
                jne     chan0441l
                jmp     channel1441
channel0441off:
channel1441:    test    dword ptr _dmacon,2       ; Channel on?
                jz      channel1441off            ; Don't make sound if dma is off
                xor     eax,eax
chan1441l:      mov     edx,4
                call    dword ptr [_audstate+edx]
                mov     ecx,dword ptr [_auddatw+4]
                mov     word ptr [_right+2*eax],cx
                inc     eax
                cmp     eax,ebx
                jne     chan1441l
                jmp     channel2441
channel1441off:
channel2441:    test    dword ptr _dmacon,4       ; Channel on?
                jz      channel2441off            ; Don't make sound if dma is off
                xor     eax,eax
chan2441l:      mov     edx,8
                call    dword ptr [_audstate+edx]
                mov     ecx,[_auddatw+8]
                add     [_right+2*eax],cx
                inc     eax
                cmp     eax,ebx
                jne     chan2441l
                jmp     channel3441
channel2441off:
channel3441:    test    dword ptr _dmacon,8       ; Channel on?
                jz      channel3441off            ; Don't make sound if dma is off
                xor     eax,eax
chan3441l:      mov     edx,12
                call    dword ptr [_audstate+edx]
                mov     ecx,dword ptr _auddatw+12
                add     word ptr [_left+2*eax],cx
                inc     eax
                cmp     eax,ebx
                jne     chan3441l
                jmp     channel4441
channel3441off:
channel4441:    jmp     dword ptr [_audio_putinbuffer]



; 33100 hz

                ALIGN32


public audio_frequency_handler_31300_:
audio_frequency_handler_31300_:
channel0313:    xor     edx,edx
                mov     dword ptr _left,edx
                mov     dword ptr _right,edx
                test    dword ptr _dmacon,1       ; Channel on?
                jz      channel0313off            ; Don't make sound if dma is off
                call    dword ptr [_audstate + edx]
                mov     eax,dword ptr _auddatw 
                mov     word ptr _left,ax
                xor     edx,edx
                call    dword ptr [_audstate + edx]
                mov     eax,dword ptr _auddatw
                mov     word ptr _left+2,ax
                jmp     channel1313
channel0313off:
channel1313:    test    dword ptr _dmacon,2       ; Channel on?
                jz      channel1313off            ; Don't make sound if dma is off
                mov     edx,4
                call    dword ptr [_audstate + edx]
                mov     eax,dword ptr _auddatw+4 
                mov     word ptr _right,ax
                mov     edx,4
                call    dword ptr [_audstate + edx]
                mov     eax,dword ptr _auddatw+4
                mov     word ptr _right+2,ax
                jmp     channel2313
channel1313off:
channel2313:    test    dword ptr _dmacon,4       ; Channel on?
                jz      channel2313off            ; Don't make sound if dma is off
                mov     edx,8
                call    dword ptr [_audstate + edx]

                mov     eax,dword ptr _auddatw+8
                add     word ptr _right,ax

                mov     edx,8
                call    dword ptr [_audstate + edx]

                mov     eax,dword ptr _auddatw+8
                add     word ptr _right+2,ax
                jmp     channel3313
channel2313off: ;or      dword ptr [_intreq],0200h  ; Testing...

channel3313:    test    dword ptr _dmacon,8       ; Channel on?
                jz      channel3313off            ; Don't make sound if dma is off
                mov     edx,12
                call    dword ptr [_audstate + edx]
                mov     eax,dword ptr _auddatw+12 
                add     word ptr _left,ax
                mov     edx,12
                call    dword ptr [_audstate + edx]
                mov     eax,dword ptr _auddatw+12
                add     word ptr _left+2,ax
                jmp     channel4313
channel3313off:
channel4313:    mov     ebx,2
                jmp     dword ptr [_audio_putinbuffer]



; 22005 hz

                ALIGN32

public audio_frequency_handler_22050_:
audio_frequency_handler_22050_:
                xor     ebx,ebx
                mov     eax,[_audiocounter]
size220:        add     eax,90020
                inc     ebx
                cmp     eax,040000h
                jbe     size220
                sub     eax,040000h
                mov     [_audiocounter],eax

                push    ebx                     ; Save number of samples

                ; Here ebx is number of samples to add

                xor     eax,eax
                mov     dword ptr [_left],eax
                mov     dword ptr [_left+2],eax
                mov     dword ptr [_right],eax
                mov     dword ptr [_right+2],eax

                test    byte ptr [_dmacon],1      ; Channel on?
                jz      channel0220off            ; No, don't make sound 
                mov     dl,byte ptr [_audioodd]
                push    edx
chan0220l:      xor     edx,edx
                call    dword ptr [_audstate + edx]
                inc     byte ptr [_audioodd]
                test    byte ptr [_audioodd],1
                jnz     chan0220dont
                mov     ecx,dword ptr [_auddatw]
                mov     word ptr [_left+2*eax],cx
                inc     eax
chan0220dont:   inc     bh
                cmp     bh,bl
                jne     chan0220l
                pop     edx
                mov     byte ptr [_audioodd],dl
channel0220off:
channel1220:    test    dword ptr _dmacon,2       ; Channel on?
                jz      channel1220off            ; Don't make sound if dma is off
                xor     bh,bh
                mov     dl,byte ptr [_audioodd]
                push    edx
                xor     eax,eax
chan1220l:      mov     edx,4
                call    dword ptr [_audstate + edx]
                inc     byte ptr [_audioodd]
                test    byte ptr [_audioodd],1
                jnz     chan1220dont
                mov     ecx,dword ptr [_auddatw+4]
                mov     word ptr [_right+2*eax],cx
                inc     eax
chan1220dont:   inc     bh
                cmp     bh,bl
                jne     chan1220l
                pop     edx
                mov     byte ptr [_audioodd],dl
channel1220off:
channel2220:    test    dword ptr _dmacon,4       ; Channel on?
                jz      channel2220off            ; Don't make sound if dma is off
                xor     bh,bh
                xor     eax,eax
                mov     dl,byte ptr [_audioodd]
                push    edx
chan2220l:      mov     edx,8
                call    dword ptr [_audstate + edx]
                inc     byte ptr [_audioodd]
                test    byte ptr [_audioodd],1
                jnz     chan2220dont
                mov     ecx,[_auddatw+8]
                add     [_right+2*eax],cx
                inc     eax
chan2220dont:   inc     bh
                cmp     bh,bl
                jne     chan2220l
                pop     edx
                mov     byte ptr [_audioodd],dl
channel2220off:
channel3220:    test    dword ptr _dmacon,8       ; Channel on?
                jz      channel3220off            ; Don't make sound if dma is off
                xor     bh,bh
                xor     eax,eax
                mov     dl,byte ptr [_audioodd]
                push    edx
chan3220l:      mov     edx,12
                call    dword ptr [_audstate + edx]
                inc     byte ptr [_audioodd]
                test    byte ptr [_audioodd],1
                jnz     chan3220dont
                mov     ecx,dword ptr _auddatw+12
                add     word ptr [_left+2*eax],cx
                inc     eax
chan3220dont:   inc     bh
                cmp     bh,bl
                jne     chan3220l
                pop     edx
                mov     byte ptr [_audioodd],dl
channel3220off:
channel4220:    pop     ebx                     ; restore number of samples
                mov     al,bl
                mov     ah,byte ptr [_audioodd]
                and     ah,1
                xor     byte ptr [_audioodd],al
                add     al,ah
                shr     al,1
                mov     bl,al
                jmp     dword ptr [_audio_putinbuffer]


; 15650

                ALIGN32

public audio_frequency_handler_15650_:
audio_frequency_handler_15650_:
channel0156:    xor     edx,edx
                mov     dword ptr _left,edx
                mov     dword ptr _right,edx
                test    dword ptr _dmacon,1       ; Channel on?
                jz      channel0156off            ; Don't make sound if dma is off
                call    dword ptr [_audstate + edx]
                mov     eax,dword ptr _auddatw 
                mov     word ptr _left,ax
                xor     edx,edx
                call    dword ptr [_audstate + edx]
                jmp     channel1156
channel0156off:
channel1156:    test    dword ptr _dmacon,2       ; Channel on?
                jz      channel1156off            ; Don't make sound if dma is off
                mov     edx,4
                call    dword ptr [_audstate + edx]
                mov     eax,dword ptr _auddatw+4 
                mov     word ptr _right,ax
                mov     edx,4
                call    dword ptr [_audstate + edx]
                jmp     channel2156
channel1156off:
channel2156:    test    dword ptr _dmacon,4       ; Channel on?
                jz      channel2156off            ; Don't make sound if dma is off
                mov     edx,8
                call    dword ptr [_audstate + edx]

                mov     eax,dword ptr _auddatw+8
                add     word ptr _right,ax

                mov     edx,8
                call    dword ptr [_audstate + edx]

                jmp     channel3156
channel2156off:
channel3156:    test    dword ptr _dmacon,8       ; Channel on?
                jz      channel3156off            ; Don't make sound if dma is off
                mov     edx,12
                call    dword ptr [_audstate + edx]
                mov     eax,dword ptr _auddatw+12 
                add     word ptr _left,ax
                mov     edx,12
                call    dword ptr [_audstate + edx]
                jmp     channel4156
channel3156off:
channel4156:    mov     ebx,1
                jmp     dword ptr [_audio_putinbuffer]

#endif

ALIGN32

_TEXT ends

_DATA segment para public 'DATA'
include data\sound.inc
include data\memory.inc
include data\inout.inc

_DATA ends

_BSS segment dword public 'DATA'
_BSS ends

end
